from functools import wraps
from itertools import chain
import inspect
import logging
import os


def debuglog(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        argspec = inspect.getargspec(func)
        argstr = dict(chain(zip(argspec.args, args), kwargs.items()))
        logging.debug('%s(%s)', func.__name__, argstr)
        try:
            retval = func(*args, **kwargs)
            logging.debug('%s -> %s)', func.__name__, retval)
            return retval
        except Exception as exc:
            logging.debug('%s => %s)', func.__name__, str(exc))
            raise

    if 'SPHERO_DEBUG' in os.environ:
        logging.basicConfig(level=logging.DEBUG)
        return wrapper
    else:
        return func



from select import select
from kulka.connection import exceptions
from kulka.core.logger import debuglog


class BaseConnection(object):

    def __init__(self, sock):
        self._sock = sock

    @classmethod
    def connect(cls, addr, port=1):
        raise NotImplementedError()

    def fileno(self):
        return self._sock.fileno()

    def close(self):
        self._sock.close()

    @debuglog
    def send(self, data, flags=0):
        retries = 10
        written = 0

        while retries > 0:
            _, wlist, _ = select([], [self._sock], [], 1)

            if wlist:
                written += wlist[0].send(data[written:], flags)

                if written >= len(data):
                    return written
            else:
                retries -= 1

        raise exceptions.ConnectionLost()

    @debuglog
    def recv(self, num_bytes):
        retries = 10

        while retries > 0:
            rlist, _, _ = select([self._sock], [], [], 1)

            if rlist:
                return rlist[0].recv(num_bytes)
            else:
                retries -= 1

        raise exceptions.ConnectionLost()



from kulka import request
from kulka.connection.exceptions import ConnectionLost
from kulka.connection import Connection
from kulka.response import parser


class Kulka(object):

    def __init__(self, addr):
        self._addr = addr
        self._connection = Connection.connect(addr)
        self._sequence = 0
        self._recv_buffer = bytearray()

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self._connection.close()

    def sequence(self):
        self._sequence = (self._sequence + 1) & 0xFF
        return self._sequence

    def _send(self, request_):
        request_.sequence = self.sequence()

        try:
            self._connection.send(request_.tobytes())
            self._wait_for_ack(request_.sequence)
        except ConnectionLost:
            self._reconnect()

    def _wait_for_ack(self, sequence):
        while True:
            try:
                response, consumed = parser(self._recv_buffer)
                self._recv_buffer = self._recv_buffer[consumed:]

                if response.seq == sequence:
                    break
            except ValueError as e:
                self._recv_buffer.extend(self._connection.recv(1024))

    def _reconnect(self):
        self._connection.close()
        self._connection = Connection.connect(self._addr)

    def set_inactivity_timeout(self, timeout):
        return self._send(request.SetInactivityTimeout(timeout))

    def set_rgb(self, red, green, blue, flag=0):
        return self._send(request.SetRGB(red, green, blue, flag))

    def roll(self, speed=0, heading=0, state=1):
        return self._send(request.Roll(speed, heading, state))

    def sleep(self, wakeup=0, macro=0, orb_basic=0):
        return self._send(request.Sleep(wakeup, macro, orb_basic))



from kulka.request.sleep import Sleep
from kulka.request.setinactivitytimeout import SetInactivityTimeout
from kulka.request.roll import Roll
from kulka.request.setrgb import SetRGB

__all__ = ['Sleep', 'SetInactivityTimeout', 'Roll', 'SetRGB']



try:
    import bluetooth
    from kulka.connection.bluezconnection import Connection
except ImportError:
    import socket

    if not hasattr(socket, 'AF_BLUETOOTH'):
        raise

    from kulka.connection.socketconnection import Connection

__all__ = ['Connection']



import struct


class Request(object):

    DID = None
    CID = None
    FMT = ''

    def __new__(cls, *_, **__):
        if cls.DID is None or cls.CID is None:
            raise NotImplementedError()

        return super(Request, cls).__new__(cls)

    def __init__(self, *args):
        self.answer = True
        self.sequence = 0
        self._data = args

    @classmethod
    def async(cls, *args):
        obj = cls(*args)
        obj.answer = False
        return obj

    def _sop(self):
        if self.answer:
            return bytearray([0xFF, 0xFF])
        else:
            return bytearray([0xFF, 0xFE])

    def tobytes(self):
        packet = bytearray([self.DID, self.CID])

        if self.answer:
            packet.append(self.sequence)

        if self._data:
            try:
                data = struct.pack(self.FMT, *self._data)
            except struct.error as exception:
                raise ValueError(str(exception))

            packet.append(len(data) + 1)
            packet.extend(data)
        else:
            packet.append(1)

        packet.append((sum(packet) & 0xFF) ^ 0xFF)

        return bytes(self._sop() + packet)



from kulka.connection.baseconnection import BaseConnection
from kulka.connection import exceptions
from kulka.core.logger import debuglog
import errno
import socket
import time


class Connection(BaseConnection):

    @classmethod
    @debuglog
    def connect(cls, addr, port=1):
        retries = 50

        while retries > 0:
            try:
                sock = socket.socket(socket.AF_BLUETOOTH,
                                     socket.SOCK_STREAM,
                                     socket.BTPROTO_RFCOMM)
                sock.connect((addr, port))
                return cls(sock)
            except socket.error as error:
                sock.close()
                time.sleep(1)

                if error.errno == errno.EHOSTDOWN:
                    retries -= 1

        raise exceptions.ConnectionFailed()



from setuptools import setup


def requirements():
    import socket

    if hasattr(socket, 'AF_BLUETOOTH'):
        return []

    return ['pybluez']


setup(
    name='Kulka',
    version='0.2',
    description='Sphero client',
    author='Karol Szuster',
    author_email='szuster.karol@gmail.com',
    url='https://github.com/karulis/pybluez',
    install_requires=requirements(),
    packages=[
        'kulka',
        'kulka.connection',
        'kulka.connection.exceptions',
        'kulka.request',
        'kulka.core'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Communications'
    ],
    download_url='https://github.com/karulis/pybluez',
    maintainer='Karol Szuster',
    license='GPL'
)



from kulka.connection import exceptions
from kulka.connection.baseconnection import BaseConnection
from kulka.core.logger import debuglog
import bluetooth
import errno
import re
import time


ERR_PATTERN = re.compile(r"\((\d+), '(.+)'\)")


@debuglog
def errstr_parse(errstr):
    match = ERR_PATTERN.match(str(errstr))

    if match is not None:
        return int(match.group(1))

    return 0


class Connection(BaseConnection):

    @classmethod
    @debuglog
    def connect(cls, addr, port=1):
        retries = 50

        while retries > 0:
            try:
                sock = bluetooth.BluetoothSocket(bluetooth.RFCOMM)
                sock.connect((addr, port))
                return cls(sock)
            except bluetooth.btcommon.BluetoothError as error:
                sock.close()
                time.sleep(1)

                if errstr_parse(error) == errno.EHOSTDOWN:
                    retries -= 1

        raise exceptions.ConnectionFailed()
